from __future__ import annotations

import warnings
from collections import defaultdict
from dataclasses import dataclass, field
from pathlib import Path
from typing import List, Dict, Iterable, Optional, Set, Mapping

from pkm.api.dependencies.dependency import Dependency
from pkm.api.environments.environment import Environment
from pkm.api.packages.package import Package
from pkm.api.repositories.repositories_configuration import RepositoryInstanceConfig, RepositoriesConfiguration, \
    RepositoriesConfigInheritanceMode
from pkm.api.repositories.repository import Repository, RepositoryBuilder, AbstractRepository
from pkm.repositories.file_repository import FileRepository
from pkm.repositories.pypi_repository import PypiRepositoryBuilder
from pkm.utils.commons import NoSuchElementException
from pkm.utils.entrypoints import EntryPoint
from pkm.utils.http.http_client import HttpClient
from pkm.utils.iterators import first_or_none

REPOSITORIES_EXTENSIONS_ENTRYPOINT_GROUP = "pkm-repositories"
REPOSITORIES_CONFIGURATION_PATH = "etc/pkm/repositories.toml"


class RepositoryLoader:
    def __init__(self, main_cfg: Path, http: HttpClient, workspace: Path):

        from pkm.api.environments.environment import Environment
        from pkm.repositories.simple_repository import SimpleRepositoryBuilder
        from pkm.repositories.git_repository import GitRepository
        from pkm.repositories.url_repository import UrlRepository
        from pkm.repositories.file_system_repository import FileSystemRepositoryBuilder

        # common builders
        pypi_builder = PypiRepositoryBuilder()

        self._builders: Dict[str, RepositoryBuilder] = {
            b.repo_type: b for b in (
                SimpleRepositoryBuilder(http),
                FileSystemRepositoryBuilder(),
                pypi_builder
            )
        }

        self._builder_providers: Dict[str, EntryPoint] = {}

        self.pypi = pypi_builder.build('pypi', {'url': 'main'})

        base_repositories = [
            GitRepository(workspace / 'git'), UrlRepository(), FileRepository()]

        # builders from entrypoints
        for epoint in Environment.current().entrypoints[REPOSITORIES_EXTENSIONS_ENTRYPOINT_GROUP]:
            try:
                ext: RepositoriesExtension = epoint.ref.import_object()()
                if not isinstance(ext, RepositoriesExtension):
                    raise ValueError("repositories entrypoint did not provide to a RepositoriesExtension class")

                for builder in ext.builders:
                    self._builders[builder.repo_type] = builder
                    self._builder_providers[builder.repo_type] = epoint

                base_repositories.extend(ext.instances)

            except Exception:  # noqa
                import traceback
                warnings.warn(f"malformed repository entrypoint: {epoint}")
                traceback.print_exc()

        base = _CompositeRepository('base', base_repositories, set(), {})

        self._cached_instances: Dict[RepositoryInstanceConfig, Repository] = {}  # noqa

        self.global_repo_config = RepositoriesConfiguration.load(main_cfg)

        self._global_repo = self.load('global', self.global_repo_config, base)
        self.workspace = workspace

    def available_repository_types(self) -> Iterable[RepositoryTypeInfo]:
        for builder in self._builders.values():
            yield RepositoryTypeInfo(builder, self._builder_providers.get(builder.repo_type))

    @property
    def global_repo(self) -> Repository:
        return self._global_repo

    def load(self, name: str, config: RepositoriesConfiguration, next_in_context: Repository) -> Repository:

        print(f"DBG: loading package config at {config.path}")

        package_search_list = []
        binding_only_repositories = set()

        for definition in config.repositories:
            instance = self.build(definition)
            package_search_list.append(instance)
            if definition.bind_only:
                binding_only_repositories.add(definition.name)

        if config.inheritance_mode == RepositoriesConfigInheritanceMode.INHERIT_CONTEXT:
            package_search_list.append(next_in_context)
        elif config.inheritance_mode == RepositoriesConfigInheritanceMode.INHERIT_GLOBAL:
            package_search_list.append(self.global_repo)

        package_binding: Dict[str, str] = {}
        print(f"DBG: {config.package_bindings=}")
        for package, binding in config.package_bindings.items():

            print(f"DBG: considering package binding: {package} = {binding}")

            if isinstance(binding, str):
                package_binding[package] = binding
            else:
                binding = self.build(RepositoryInstanceConfig.from_config(f'_unnamed_repo_for_{package}', binding))
                print(f"DBG: build unnamed repository: {binding.name}")
                binding_only_repositories.add(binding.name)
                package_search_list.append(binding)
                package_binding[package] = binding.name

        if not package_search_list:
            if config.inheritance_mode == RepositoriesConfigInheritanceMode.INHERIT_GLOBAL:
                return self.global_repo
            return next_in_context

        return _CompositeRepository(name, package_search_list, binding_only_repositories, package_binding)

    def build(self, config: RepositoryInstanceConfig) -> Repository:
        if not (cached := self._cached_instances.get(config)):
            if not (builder := self._builders.get(config.type)):
                raise KeyError(f"unknown repository type: {config.type}")
            cached = builder.build(config.name, config.args)
            self._cached_instances[config] = cached

        return cached


class _CompositeRepository(AbstractRepository):
    def __init__(
            self, name: str, package_search_list: List[Repository],
            binding_only_repositories: Set[str], package_bindings: Mapping[str, str]):
        super().__init__(name)

        self._url_handlers: Dict[str, List[Repository]] = defaultdict(list)
        self._package_search_list = []
        self._binding_only_repositories = binding_only_repositories
        self._package_binding = package_bindings

        for repo in package_search_list:
            if repo.accept_non_url_packages():
                self._package_search_list.append(repo)
            for protocol in repo.accepted_url_protocols():
                self._url_handlers[protocol].append(repo)

    def _do_match(self, dependency: Dependency, env: Environment) -> List[Package]:

        if url := dependency.required_url():
            for repo in self._url_handlers[url.protocol]:
                if match := repo.match(dependency, env):
                    return match
            raise []

        if repo_name := self._package_binding.get(dependency.package_name):
            if not (repo := first_or_none(it for it in self._package_search_list if it.name == repo_name)):
                print(f"DBG: repositories are {[r.name for r in self._package_search_list]}")
                print(f"DBG: type of repo_name is {type(repo_name)}")

                raise NoSuchElementException(f"package: {dependency.package_name} is bound to "
                                             f"repository: {repo_name}, but this repository is not defined")
            return repo.match(dependency, env)

        for repo in self._package_search_list:
            if repo.name not in self._binding_only_repositories and (result := repo.match(dependency, env)):
                return result

        return []

    def _sort_by_priority(self, dependency: Dependency, packages: List[Package]) -> List[Package]:
        return packages

    def accepted_url_protocols(self) -> Iterable[str]:
        return self._url_handlers.keys()


@dataclass
class RepositoryTypeInfo:
    builder: RepositoryBuilder
    provider: Optional[EntryPoint]

    def is_builtin(self):
        return self.provider is None

    @property
    def type_name(self) -> str:
        return self.builder.repo_type


@dataclass
class RepositoriesExtension:
    builders: List[RepositoryBuilder] = field(default_factory=list)
    instances: List[Repository] = field(default_factory=list)



from typing import TypeVar, Set

from pkm.utils.types import Hashable

_T = TypeVar("_T", bound=Hashable)


def try_add(st: Set[_T], item: _T) -> bool:
    """
    tries to add an item to the set indicating if the operation succeeded
    :param st: the set to add the item into
    :param item: the item to add
    :return: True if it was added, False if it was already in the set
    """

    ln = len(st)
    st.add(item)
    return ln != len(st)



from __future__ import annotations

import warnings
from copy import copy
from dataclasses import dataclass
from enum import Enum
from typing import Optional, Any, Dict, List, Mapping

from pkm.config.configuration import TomlFileConfiguration, computed_based_on
from pkm.utils.dicts import remove_none_values
from pkm.utils.hashes import HashBuilder


class RepositoriesConfigInheritanceMode(Enum):
    INHERIT_CONTEXT = 0,
    INHERIT_GLOBAL = 1,
    NO_INHERITANCE = 2


class RepositoriesConfiguration(TomlFileConfiguration):
    repositories: List[RepositoryInstanceConfig]
    inheritance_mode: RepositoriesConfigInheritanceMode
    package_bindings: Mapping[str, Any]

    @computed_based_on("inheritance")
    def inheritance_mode(self) -> RepositoriesConfigInheritanceMode:
        value = self["inheritance"]
        if value is not None:
            if value == "global":
                return RepositoriesConfigInheritanceMode.INHERIT_GLOBAL
            if not value:
                return RepositoriesConfigInheritanceMode.NO_INHERITANCE
        return RepositoriesConfigInheritanceMode.INHERIT_CONTEXT

    @computed_based_on("package-bindings")
    def package_bindings(self) -> Mapping[str, Any]:
        return self["package-bindings"] or {}

    @package_bindings.modifier
    def set_package_bindings(self, new_bindings: Mapping[str, Any]):
        self["package-bindings"] = new_bindings

    @computed_based_on("repos")
    def repositories(self) -> List[RepositoryInstanceConfig]:
        repositories: Dict[str, Dict] = self["repos"] or {}

        def read_config_warn_err(name: str, repo: Dict[str, str]) -> Optional[RepositoryInstanceConfig]:
            try:
                return RepositoryInstanceConfig.from_config(name, repo)
            except Exception as e:
                warnings.warn(f"could not process repository config for {name}: {e}, ignoring its configuration")
                return None

        return [
            repo_instance for name, repo in repositories.items()
            if (repo_instance := read_config_warn_err(name, repo)) is not None]

    @repositories.modifier
    def set_repositories(self, repos: List[RepositoryInstanceConfig]):
        self["repos"] = {r.name: r.to_config() for r in repos}


@dataclass(frozen=True, eq=True)
class RepositoryInstanceConfig:
    type: str
    bind_only: Optional[bool]
    name: Optional[str]
    args: Dict[str, str]

    def __hash__(self):
        return HashBuilder() \
            .regulars(self.type, self.name, self.bind_only) \
            .unordered_mapping(self.args) \
            .build()

    def to_config(self) -> Dict[str, Any]:
        return remove_none_values({
            **self.args,
            'type': self.type,
            'bind-only': self.bind_only
        })

    @classmethod
    def from_config(cls, name: str, config: Dict[str, Any]) -> "RepositoryInstanceConfig":
        config = copy(config)
        type_ = config.pop('type')
        bind_only: bool = config.pop('bind-only', False)
        args = config
        return RepositoryInstanceConfig(type_, bind_only, name, args)



from collections import defaultdict
from pathlib import Path
from typing import List, Dict, Optional

from pkm.api.dependencies.dependency import Dependency
from pkm.api.distributions.distribution import Distribution
from pkm.api.environments.environment import Environment
from pkm.api.packages.package import Package
from pkm.api.projects.project import Project
from pkm.api.projects.project_group import ProjectGroup
from pkm.api.repositories.repository import AbstractRepository, RepositoryBuilder, Repository
from pkm.resolution.pubgrub import MalformedPackageException
from pkm.utils.strings import endswith_any

FILE_SYSTEM_REPOSITORY_TYPE = "file-system"


class PackagesDictRepository(AbstractRepository):

    def __init__(self, name: str, packages: Dict[str, List[Package]]):
        super().__init__(name)
        self._packages = packages

    def _do_match(self, dependency: Dependency, env: Environment) -> List[Package]:
        all_packages = self._packages.get(dependency.package_name) or []
        return [p for p in all_packages if dependency.version_spec.allows_version(p.version)]


class FileSystemRepositoryBuilder(RepositoryBuilder):

    def __init__(self):
        super().__init__(FILE_SYSTEM_REPOSITORY_TYPE)

    def build(self, name: Optional[str], args: Dict[str, str]) -> Repository:
        path = Path(self._arg(args, 'path', required=True))

        packages: Dict[str, List[Package]] = defaultdict(list)
        if path.is_dir():
            if ProjectGroup.is_valid(path):
                project_group = ProjectGroup.load(path)
                for project in project_group.project_children_recursive:
                    packages[project.name].append(project)
            else:
                try:
                    project = Project.load(path)
                    packages[project.name].append(project)
                except MalformedPackageException:
                    pass

            for dist in path.iterdir():
                if _is_distribution(dist):
                    package = Distribution.package_from(dist)
                    packages[package.name].append(package)
        elif _is_distribution(path):
            package = Distribution.package_from(path)
            packages[package.name].append(package)

        return PackagesDictRepository(name, packages)


def _is_distribution(path: Path) -> bool:
    return endswith_any(path.name, ('.tar.gz', '.whl')) is not None



from base64 import b64encode
from dataclasses import dataclass
from typing import Tuple


@dataclass(frozen=True, eq=True)
class BasicAuthentication:
    username: str
    password: str

    def as_header(self) -> Tuple[str, str]:
        return 'Authorization', f'Basic {b64encode(f"{self.username}:{self.password}".encode()).decode("ascii")}'



from __future__ import annotations
from abc import ABC, abstractmethod
from pathlib import Path
from typing import Dict, List, Optional, Tuple, Iterable, Union

from pkm.api.dependencies.dependency import Dependency
from pkm.api.environments.environment import Environment
from pkm.api.environments.environments_zoo import EnvironmentsZoo
from pkm.api.packages.package import PackageDescriptor, Package
from pkm.api.pkm import HasAttachedRepository, Pkm, pkm
from pkm.api.projects.project import Project
from pkm.api.projects.project_group import ProjectGroup
from pkm.api.repositories.repositories_configuration import RepositoriesConfiguration, RepositoryInstanceConfig
from pkm.api.repositories.repository import Repository, AbstractRepository, RepositoryPublisher
from pkm.api.repositories.repository_loader import RepositoryLoader, REPOSITORIES_CONFIGURATION_PATH
from pkm.repositories.shared_pacakges_repo import SharedPackagesRepository
from pkm.resolution.packages_lock import LockPrioritizingRepository
from pkm.utils.commons import NoSuchElementException
from pkm.utils.properties import cached_property, clear_cached_properties
from pkm.utils.sequences import pop_or_none
from pkm.utils.sets import try_add


class RepositoryManagement(ABC):
    def __init__(self, cfg: RepositoriesConfiguration, loader: Optional[RepositoryLoader] = None):
        if not loader:
            loader = pkm.repository_loader

        self._loader = loader
        self.configuration = cfg

    @abstractmethod
    def _load_attached(self) -> Repository:
        ...

    @cached_property
    def _publishers(self) -> Dict[str, Optional[RepositoryPublisher]]:
        return {}  # it is cached only so that _update_config will clear it

    @abstractmethod
    def parent_contexts(self) -> List[HasAttachedRepository]:
        ...

    def publisher_for(self, name: str) -> Optional[RepositoryPublisher]:
        if name in self._publishers:
            return self._publishers[name]

        contexes = [self]
        opened = set()
        while context := pop_or_none(contexes):
            if not try_add(opened, id(context)):
                continue
            contexes.extend(context.parent_contexts())
            for repo_config in context.defined_repositories():
                if repo_config.name == name:
                    result = self._publishers[name] = self._loader.build(repo_config).publisher
                    return result

        raise NoSuchElementException(f"repository: {name} could not be found")

    def register_bindings(self, packages: List[str], repo: Optional[Union[str, RepositoryInstanceConfig]]):
        if repo:
            if isinstance(repo, RepositoryInstanceConfig):
                repo = repo.to_config()

            self.configuration.package_bindings = {
                **self.configuration.package_bindings,
                **{p: repo for p in packages}
            }
        else:
            self.configuration.package_bindings = {
                p: r for p, r in self.configuration.package_bindings.items()
                if p not in packages
            }

        self._update_config()

    @cached_property
    def attached_repo(self) -> Repository:
        return self._load_attached()

    def _update_config(self):
        self.configuration.save()
        clear_cached_properties(self)

    def add_repository(self, name: str, builder: str, args: Dict[str, str], bind_only: bool = False):
        config = RepositoryInstanceConfig(builder, bind_only, name, args)

        # the following line act a s a safeguard before the add operation,
        # if it fails the actual addition will not take place
        self._loader.build(config)

        self.configuration.repositories = [
            *self.configuration.repositories, config]

        self._update_config()

    def defined_repositories(self) -> List[RepositoryInstanceConfig]:
        return self.configuration.repositories

    def remove_repository(self, name: str):
        self.configuration.repositories = [repo for repo in self.configuration.repositories if repo.name != name]
        self.configuration.package_bindings = {
            p: r for p, r in self.configuration.package_bindings.items() if r != name}
        self._update_config()


def _config_at(path: Path) -> RepositoriesConfiguration:
    return RepositoriesConfiguration.load(path / REPOSITORIES_CONFIGURATION_PATH)


class PkmRepositoryManagement(RepositoryManagement):

    def __init__(self, pkm_: Pkm):
        super().__init__(pkm_.repository_loader.global_repo_config, pkm_.repository_loader)
        self.pkm = pkm_

    def _load_attached(self) -> Repository:
        return self._loader.global_repo

    def parent_contexts(self) -> List[HasAttachedRepository]:
        return []


class EnvRepositoryManagement(RepositoryManagement):

    def __init__(self, env: Environment, loader: Optional[RepositoryLoader] = None):
        super().__init__(_config_at(env.path), loader)
        self.env = env

    def parent_contexts(self) -> List[HasAttachedRepository]:
        return [self.env.zoo or pkm]

    def _load_attached(self) -> Repository:
        repo = self._loader.global_repo
        if zoo := self.env.zoo:
            repo = zoo.attached_repository

        if self.configuration.path.exists():
            repo = self._loader.load("env-configured-repository", self.configuration, repo)

        return repo


class ZooRepositoryManagement(RepositoryManagement):

    def __init__(self, zoo: EnvironmentsZoo, loader: Optional[RepositoryLoader] = None):
        super().__init__(_config_at(zoo.path), loader)
        self.zoo = zoo

    def parent_contexts(self) -> List[HasAttachedRepository]:
        return pkm

    def _load_attached(self) -> Repository:
        repo = self._loader.global_repo
        if self.configuration.path.exists():
            repo = self._loader.load("zoo-configured-repository", self.configuration, repo)

        if self.zoo.config.package_sharing.enabled:
            repo = SharedPackagesRepository(self.zoo.path / ".zoo/shared", repo)

        return repo


def _load_for_project_group(
        loader: RepositoryLoader,
        group: ProjectGroup,
        config: RepositoriesConfiguration,
        base_repo: Optional[Repository] = None) -> Repository:
    repo = base_repo or loader.global_repo

    if config.path.exists():
        repo = loader.load("group-configured-repository", config, repo)

    repo = _ProjectsRepository.create("group-projects-repository", group.project_children_recursive, repo)
    return repo


class ProjectRepositoryManagement(RepositoryManagement):

    def __init__(self, project: Project, loader: Optional[RepositoryLoader] = None):
        super().__init__(_config_at(project.path), loader)
        self.project = project

    def parent_contexts(self) -> List[HasAttachedRepository]:
        result = [self.project.attached_environment]
        if self.project.group:
            result.append(self.project.group)

        return result

    def _load_attached(self) -> Repository:
        repo = self.project.attached_environment.attached_repository

        if group := self.project.group:
            repo = _load_for_project_group(self._loader, group, _config_at(group.path), repo)
        else:
            repo = _ProjectsRepository.create('project-repository', [self.project], repo)

        if self.configuration.exists():
            repo = self._loader.load("project-configured-repository", self.configuration, repo)

        repo = LockPrioritizingRepository(
            "lock-prioritizing-repository", repo, self.project.lock,
            self.project.attached_environment)

        return repo


class ProjectGroupRepositoryManagement(RepositoryManagement):

    def __init__(self, group: ProjectGroup, loader: Optional[RepositoryLoader] = None):
        super().__init__(_config_at(group.path), loader)
        self.group = group

    def _load_attached(self) -> Repository:
        return _load_for_project_group(self._loader, self.group, self.configuration, None)

    def parent_contexts(self) -> List[HasAttachedRepository]:
        return [pkm]


class _ProjectsRepository(AbstractRepository):
    def __init__(self, name: str, projects: Dict[str, Tuple[PackageDescriptor, Path]], base_repo: Repository):
        super().__init__(name)
        self._packages = projects
        self._base_repo = base_repo

    def _do_match(self, dependency: Dependency, env: Environment) -> List[Package]:
        if (package_and_path := self._packages.get(dependency.package_name)) and \
                dependency.version_spec.allows_version(package_and_path[0].version):
            return [Project.load(package_and_path[1])]
        return self._base_repo.match(dependency, env)

    def _sort_by_priority(self, dependency: Dependency, packages: List[Package]) -> List[Package]:
        return packages

    @classmethod
    def create(cls, name: str, projects: Iterable[Project], base: Repository) -> _ProjectsRepository:
        return _ProjectsRepository(name, {p.name: (p.descriptor, p.path) for p in projects}, base)

    def accept_non_url_packages(self) -> bool:
        return self._base_repo.accept_non_url_packages()

    def accepted_url_protocols(self) -> Iterable[str]:
        return self._base_repo.accepted_url_protocols()
